"""
json_serializable
~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`JsonSerializable` along with supporting classes,
functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import json
from abc import ABC, abstractmethod
from io import TextIOBase
from typing import Union

from inception_tools.constants import UNIMPLEMENTED_ABSTRACT_METHOD_ERROR
from inception_tools.serializable import Serializable, SerializationError

JSON_OBJ_TYPE = Union[dict, list, str, int, float, bool]
"""
The type definition for the ``fp`` argument in
:py:meth:`JsonSerializable.from_json` and
:py:meth:`JsonSerializable.to_json`.
"""


class JsonSerializationError(SerializationError):
    """
    A common exception raised by subclasses of :py:class:`JsonSerializable`.
    """


class JsonSerializable(Serializable, ABC):
    """
    Provides an interface by which classes can make themselves
    :py:class:`Serializable` by providing a way to marshal and unmarshal themselves
    from and to a JSON-like Python object representation.  The JSON-like Python
    object representation can be anything accepted by the :py:func:`json.dump` or
    produced by :py:func:`json.load`, i.e., a composition of objects of the following
    types:

    - ``dict``
    - ``list``
    - ``str``
    - ``bool``
    - ``int``
    - ``float``
    - ``None``
    """

    ##############################
    # Static / class methods

    @classmethod
    def from_text_io(cls, fp: TextIOBase):
        """
        Deserializes a JSON-formatted text-based input stream by first converting it
        to a JSON-like Python object using the :py:mod:`json` package and delegating
        responsibility for unmarshalling the JSON representation to
        :py:meth:`from_json`.
        """
        json_obj = json.load(fp)
        return cls.from_json(json_obj)

    @classmethod
    @abstractmethod
    def from_json(cls, json_obj: JSON_OBJ_TYPE):
        """
        Unmarshals a new ``cls`` instance from JSON-like Python object.
        :param json_obj: the JSON-like Python object
        :return: the new ``cls`` instance
        :raises SerializationError: if a problem occurs during execution of this method
        .. seealso:: modules :py:class:`JsonSerializable`
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    ##############################
    # Static / class methods

    def to_text_io(self, fp: TextIOBase):
        """
        Serializes this instance to a JSON-formatted text-based output stream by
        first converting it to a JSON-like Python object using the :py:mod:`json`
        package and delegating responsibility for serializing the JSON representation
        to the :py:mod:`json` package.
        """
        json_obj = self.to_json()
        return json.dump(json_obj, fp)

    def to_json(self) -> JSON_OBJ_TYPE:
        """
        Marshals a new ``cls`` instance to JSON-like Python object.
        :return: the JSON-like Python object
        :raises SerializationError: if a problem occurs during execution of this method
        .. seealso:: modules :py:class:`JsonSerializable`
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR



"""
archetype_metadata
~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`ArchetypeMetadata` along with supporting
classes, functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from collections import namedtuple

from jsonschema import validate, ValidationError

from inception_tools.json_serializable import (
    JSON_OBJ_TYPE,
    JsonSerializable,
    JsonSerializationError,
)


class ArchetypeMetadata(
    namedtuple(
        "ArchetypeMetadataBase",
        (
            "group_id",
            "archetype_id",
            "version_id",
        ),
    ),
    JsonSerializable,
):
    """
    Instances of this class encapsulate and group metadata that pertains to an
    :py:class:`archetype.Archetype`.  :py:class:`Archetype` instances are uniquely
    identified by the combination of a group identifier (e.g., 'inception_tools'),
    which determines a name space, an architype identifier ( e.g., standard),
    and a version identifier ( e.g., '1.2.3').  The combination of architype and
    version identifier is assumed to be unique within the name space of the group
    identifier. :ivar str group_id: the group identifier :ivar str archetype_id: the
    architype identifier :ivar str version_id: the architype version identifier
    """

    # Make instances of this class immutable
    __slots__ = ()

    JSON_SCHEMA = {
        "type": "object",
        "properties": {
            "group_id": {"type": "string"},
            "archetype_id": {"type": "string"},
            "version_id": {"type": "string"},
        },
    }
    """
    The Python `JSON schema`_ that :py:meth:`serialize_json` and
    :py:meth:`deserialize_json` must adhere to.

    .. _`JSON schema`: https://json-schema.org/
    """

    @classmethod
    def _validate_json(cls, json_obj: JSON_OBJ_TYPE) -> None:
        try:
            validate(json_obj, cls.JSON_SCHEMA)
        except ValidationError as e:
            to_raise = JsonSerializationError(f"Invalid JSON format: {json_obj}")
            raise to_raise from e

    @classmethod
    def from_json(cls, json_obj: JSON_OBJ_TYPE):
        """
        Creates a new :py:class:`ArchetypeMetadata` from a JSON-like Python object of
        the form:

        .. code-block::

            {
                'group_id': <group-id-string>,
                'archetype_id': <archetype-id-string>,
                'version_id': <version-id-string>,
            }
        """
        cls._validate_json(json_obj)

        if isinstance(json_obj, dict):
            return ArchetypeMetadata(
                group_id=json_obj["group_id"],
                archetype_id=json_obj["archetype_id"],
                version_id=json_obj["version_id"],
            )

        raise RuntimeError(
            f"Expected an object of type 'dict' but received: {json_obj!r}"
        )

    def to_json(self) -> dict:
        """
        Returns a JSON-like Python object of the form:

        .. code-block::

            {
                'group_id': <group-id-string>,
                'archetype_id': <archetype-id-string>,
                'version_id': <version-id-string>,
            }
        """
        json_obj = self._asdict()
        self._validate_json(json_obj)
        return json_obj



"""
template_file_builder
~~~~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`TemplateFileBuilder` along with
supporting classes, functions, and attributes.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import os

from jinja2 import Template

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.file_builder import FileBuilder


class TemplateFileBuilder(FileBuilder):
    """
    This class uses :py:class:jinja2.Template` instances to create both the subpath
    and content of a file.
    """

    PATH_SEP = "/"
    """
    The path separator to use to write ``subpath`` templates. This separator will be
    converted to the OS-specific path separator automatically by :py:meth:`subpath`.
    """

    @classmethod
    def from_strings(cls, subpath: str, prototype: str) -> FileBuilder:
        """
        Factory method that builds a new :py:class:`TemplateDirectoryBuilder`
        instance from :py:class:`jinja2.Template` source ``string``s, which will be
        used to create the return value of :py:meth:`subpath` and :py:meth:`render`
        from the ``params`` argument.
        :param subpath: the subpath template string
        :param prototype: the prototype template string
        :return: the new instance
        .. seealso:: :py:meth:`__init__`
        """
        s = Template(subpath)
        p = Template(prototype, keep_trailing_newline=True)
        return cls(s, p)

    def __init__(self, subpath: Template, prototype: Template) -> None:
        """
        Initializes a new :py:class:`TemplateFileBuilder` instance.
        :param subpath: the template used to produce the return value of
        :py:meth:`subpath`
        :param prototype: the template used to produce the return value of
        :py:meth:`subpath`
        .. seealso:: :py:attr:`PATH_SET`, :py:meth:`subpath`, :py:meth:`render`
        """
        super().__init__()
        self._subpath = subpath
        self._prototype = prototype

    def subpath(self, params: ArchetypeParameters) -> str:
        """
        Creates the subpath using the ``subpath`` :py:class:`jinja2.Template` used to
        initialize this instance, using the named :py:class:`ArchetypeParameters` to
        replace any template variables.  The path, once rendered, is split using
        :py:attr:`PATH_SEP` and rejoined using the OS-specific path separator.
        """
        subpath_raw = self._subpath.render(**params.as_dict())
        subpath_split = subpath_raw.split(self.PATH_SEP)
        return os.path.join(*subpath_split)

    def render(self, params: ArchetypeParameters) -> str:
        """
        Renders the file content using the ``subpath`` :py:class:`jinja2.Template`
        used to initialize this instance, using the named
        :py:class:`ArchetypeParameters` to replace any template variables.
        """
        return self._prototype.render(**params.as_dict())



"""
template_directory_builder
~~~~~~~~~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`TemplateDirectoryBuilder` along with supporting
classes, functions, and attributes.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import os

from jinja2 import Template

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.directory_builder import DirectoryBuilder


class TemplateDirectoryBuilder(DirectoryBuilder):
    """
    An implementation of :py:class:`inception_tools.DirectoryBuilder` which uses a
    :py:class:`jinja2.Template` to create the subpath for the directory to be built.
    """

    PATH_SEP = "/"
    """
    The path separator to use to write ``subpath`` templates. This separator will be
    converted to the OS-specific path separator automatically by :py:meth:`subpath`.
    """

    @classmethod
    def from_string(cls, subpath: str) -> DirectoryBuilder:
        """
        Factory method that builds a new :py:class:`TemplateDirectoryBuilder`
        instance from a :py:class:`jinja2.Template` source ``string``, which will be
        used to create the return value of :py:meth:`subpath` from the ``params``
        argument.

        :param subpath: the subpath template string
        :return: the new instance
        .. seealso:: :py:meth:`__init__`
        """
        t = Template(subpath)
        return cls(t)

    def __init__(self, template: Template) -> None:
        """
        Initializes a new :py:class:`TemplateDirectoryBuilder` instance. :param
        template: a :py:class:`jinja2.Template` to be used with the ``params``
        argument of :py:meth:`subpath`.
        """
        self._template = template

    def subpath(self, params: ArchetypeParameters) -> str:
        """
        Renders the template held by using ``params`` dictionary representation.
        """
        subpath_raw = self._template.render(**params.as_dict())
        subpath_split = subpath_raw.split(self.PATH_SEP)
        return os.path.join(*subpath_split)



"""
serializable
~~~~~~~~~~~~

Houses the declaration of :py:class:`Serializable` along with supporting classes,
functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from abc import ABC, abstractmethod
from contextlib import closing
from io import StringIO, TextIOBase
from typing import Union

from inception_tools.constants import UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

TEXT_IO_TYPE = Union[TextIOBase, StringIO]


class SerializationError(Exception):
    """
    A common exception type for subclasses of :py:class:`Serializable` may raise to
    indicate an error.
    """


class Serializable(ABC):
    """
    Defines the common interface by which objects are serialized and deserialized
    within the :py:mod:`inception_tools` package.
    """

    ##############################
    # Static / class methods

    @classmethod
    @abstractmethod
    def from_text_io(cls, file_obj: TEXT_IO_TYPE):
        """
        Deserializes a new ``cls`` instance from a text-based input stream.
        :param file_obj: the file-like object
        :return: a new instance of type ``cls``
        :raises SerializationError: if a problem occurs during execution of this method
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    @classmethod
    def from_string(cls, s):
        """
        Deserializes a new ``cls`` instance from a ``str``.
        :param s: the ``str``
        :return: a new instance of type ``cls``
        :raises SerializationError: if a problem occurs during execution of this method
        """
        with closing(StringIO(s)) as f:
            return cls.from_text_io(f)

    ##############################
    # Instance methods

    @abstractmethod
    def to_text_io(self, file_obj: TEXT_IO_TYPE):
        """
        Serializes this instance to a text-based output stream.
        :param file_obj: the input stream
        :raises SerializationError: if a problem occurs during execution of this method
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    def to_string(self) -> str:
        """
        Serializes this instance to a ``str``.
        :return: the ``str``
        :raises SerializationError: if a problem occurs during execution of this method
        """
        with closing(StringIO()) as f:
            self.to_text_io(f)
            return f.getvalue()



"""
main
~~~~~

Main entry point commend line script.  See :py:func:`cli` for details of how to
invoke this script.
"""

import logging
import pathlib
from datetime import datetime
from logging.config import fileConfig

import click

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.exception import LoggingConfigError
from inception_tools.standard_archetype import StandardArchetype

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"


def _logger() -> logging.Logger:
    return logging.getLogger(__file__)


def _incept(
    package_name: str,
    project_root: str,
    author: str,
    author_email: str,
    archetype_name: str,
) -> None:
    archetype = StandardArchetype.from_string(archetype_name)
    params = ArchetypeParameters(
        package_name=package_name,
        author=author,
        author_email=author_email,
        date=datetime.now(),
    )
    root_dir = project_root or package_name
    archetype.build(root_dir=root_dir, params=params)


@click.command()
@click.argument("package_name")
@click.argument(
    "project_root",
    type=click.Path(file_okay=False),
    default=None,
    required=False,
)
@click.option(
    "--author-name",
    type=str,
    default="[insert-author-name]",
    help="The name of the package author. Depending on the archetype selected, this "
    "may be used in copyright notices, setup.py package metadata, and for "
    "__author__ assignments in Python stub files.",
)
@click.option(
    "--author-email",
    type=str,
    default="[insert-author-email]",
    help="The email of the package author. Depending on the archetype selected, "
    "this bay be used in setup.py package metadata and in the auto-generated "
    "boilerplate text of README.rst file.",
)
@click.option(
    "--archetype",
    default=StandardArchetype.CLI.canonical_name,
    type=click.Choice(StandardArchetype.canonical_names(), case_sensitive=False),
)
def incept(
    package_name: str,
    project_root: str,
    author_name: str,
    author_email: str,
    archetype: str,
) -> None:
    """
    Builds a new project structure with the given package name.  Command line
    syntax:

        it incept <package-name>

    Invoking the command line above will _result in the creation of a directory with
    the following structure:

    \b
        <working-dir>/
            my_package/
                my_package/
                    __init__.py
                    my_package.py
                tests/
                    __init__.py
                    end-to-end/
                        __init__.py
                        test_my_package/
                            __init__.py
                    integration/
                        __init__.py
                        test_my_package/
                            __init__.py
                    unit/
                        __init__.py
                        test_my_package/
                            __init__.py
                LICENSE
                Makefile
                Pipfile
                README.rst
                setup.cfg
                setup.py

    where many of the files are parameterized with the package name, author name,
    author email, etc.

    PACKAGE_NAME: the name of the package to be created.  This should be the package
    name, as you would expect it to appear in code references ( e.g. 'my_package' and
    not 'my-package'

    PROJECT_ROOT (optional - defaults to PACKAGE_NAME): the directory under which the
    new project should be created. If the directory does not already exist, it
    will be created automatically. Any preexisting files and/or subdirectories matching
    those created by this command will be overwritten.
    """
    try:
        _incept(package_name, project_root, author_name, author_email, archetype)
    except Exception:
        msg = (
            f"Unexpected exception: "
            f"package_name={package_name!r}, project_root={project_root!r}, "
            f"author_name={author_name!r}, author_email={author_email!r}"
        )
        _logger().exception(msg)
        raise


@click.group()
@click.option(
    "-l", "--logging-config", default=None, type=click.Path(exists=True, dir_okay=False)
)
def cli(logging_config: str) -> None:
    """
    Main command-line application for the inception_tools package.  This application
    can be used to access various commands listed below.  For example to incept a new
    project called 'my_package', use the following command:

        it incept my_package

    For additional help using any command, use the help for the command as follows

        it <command> --help
    """
    logging_config = logging_config or "log.cfg"
    if pathlib.Path(logging_config).is_file():
        try:
            fileConfig(logging_config, disable_existing_loggers=False)
        except Exception:
            raise LoggingConfigError(
                f"Could not parse the logging config file: {logging_config!r}"
            )
    else:
        logging.basicConfig(level=logging.INFO)
        _logger().warning(
            f"Logging config file not found: {logging_config!r}."
            "All logs will be directed to a basic console logger."
        )
    _logger().debug("Successfully set up logging.")


cli.add_command(incept)

if __name__ == "__main__":
    cli()



"""
    __init__.py
    ~~~~~~~~~~~

    The :py:mod:`inception_tools` package contains the root-level distributable
    source code for the `inception_tools` project.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("inception_tools").version
except pkg_resources.DistributionNotFound:
    __version__ = "unknown"



"""
archetype_parameters
~~~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`ArchetypeParameters` along with supporting
classes, functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from collections import namedtuple


class ArchetypeParameters(
    namedtuple(
        "ArchetypeParametersBase", ("package_name", "author", "author_email", "date")
    )
):
    """
    A container class that responsible for grouping the parameters used to create a
    project directory structure from an :py:class:`archetype.Archetype`.

    Instances of this class are immutable.

    :ivar str package_name: the name of the package to be created
    :ivar str author: the package author name
    :ivar str author_email: the author's email address
    :ivar str date: the inception date of the project
    """

    def as_dict(self) -> dict:
        """
        Returns a dictionary representation of this instance, wherein each named
        attribute is a key in the dictionary returned.
        :return: the dictionary representing this instance
        """
        return self._asdict()



"""
archetype_base
~~~~~~~~~~~~~~

Houses the declaration of :py:class:`ArchetypeBase` along with supporting classes,
functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from typing import Iterable

from inception_tools.archetype import Archetype
from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.directory_builder import DirectoryBuilder
from inception_tools.file_builder import FileBuilder


class ArchetypeBase(Archetype):
    """
    A base implementation of :py:class:`Archetype` that provides basic
    implementations of :py:meth:`Archetype.build` and :py:meth:`Archetype.file_paths`.
    """

    def __init__(
        self,
        file_builders: Iterable[FileBuilder],
        dir_builders: Iterable[DirectoryBuilder],
    ) -> None:
        """
        Class initializer.
        :param file_builders: a set of
        :py:class:`inception_tools.ArchetypeResourceBuilder` instances used by
        :py:meth:`build` to create the project structure.
        """
        super().__init__()
        self._file_builders = file_builders
        self._dir_builders = dir_builders

    def file_paths(self, root_path: str, params: ArchetypeParameters) -> Iterable[str]:
        return tuple(r.path(root_path, params) for r in self._file_builders)

    def dir_paths(self, root_path: str, params: ArchetypeParameters) -> Iterable[str]:
        return tuple(r.path(root_path, params) for r in self._dir_builders)

    def build(self, root_dir: str, params: ArchetypeParameters) -> None:
        """
        Builds the project structure using the :py:class:`FileBuilder` instances held
        by this instance.
        :param root_dir: the root directory of the project structure to be
        created
        :param params: the :py:class:`ArchetypeParameters` to use as context
        for the project to be built
        :return: :py:const:`None`
        """
        for r in self._file_builders:
            r.build(root_dir, params)
        for r in self._dir_builders:
            r.build(root_dir, params)



"""
standard_archetype
~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`StandardArchetype` along with supporting
classes, functions, and attributes.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import os
from abc import ABCMeta
from enum import Enum, EnumMeta
from typing import Iterable, Tuple

from inception_tools.archetype import Archetype
from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.template_archetype import TemplateArchetype

ARCHETYPE_DIR = os.path.abspath(os.path.join(__file__, os.pardir, "data", "archetypes"))


class _ABCEnumMeta(ABCMeta, EnumMeta):
    # Enables Enums to inherit from abstract base classes
    pass


class StandardArchetype(Archetype, Enum, metaclass=_ABCEnumMeta):
    """
    Enumerates the standard :py:class:`Archetype` instances available across the
    system.
    """

    CLI = ("inception-tools-archetype-cli-1.0",)
    """
    The :py:meth:`build` method of this :py:class:`Archetype` will create a
    directory/file tree with the following structure:

    ::

        root_dir/
            my_package/
                __init__.py
                cli.py
            tests/
                __init__.py
                end-to-end/
                    __init__.py
                integration/
                    __init__.py
                unit/
                    __init__.py
            LICENSE
            Makefile
            Pipfile
            README.rst
            setup.cfg
            setup.py

    where 'root_dir' is the `root_dir argument and 'my_package' is the `package_name`
    attribute of the params argument.
    """

    LIB = ("inception-tools-archetype-lib-1.0",)
    """
    The :py:meth:`build` method of this :py:class:`Archetype` will create a
    directory/file tree with the following structure:

    ::

        root_dir/
            my_package/
                __init__.py
                my_package.py
            tests/
                __init__.py
                end-to-end/
                    __init__.py
                integration/
                    __init__.py
                unit/
                    __init__.py
            LICENSE
            Makefile
            Pipfile
            README.rst
            setup.cfg
            setup.py

    where 'root_dir' is the `root_dir argument and 'my_package' is the `package_name`
    attribute of the params argument.
    """

    SIMPLE = ("inception-tools-archetype-simple-1.0",)
    """
    The :py:meth:`build` method of this :py:class:`Archetype` will create a
    directory/file tree with the following structure:

    ::

        root_dir/
            my_package.py
            tests/
                __init__.py
            LICENSE
            Makefile
            Pipfile
            README.rst
            setup.cfg
            setup.py

    where 'root_dir' is the `root_dir argument and 'my_package' is the `package_name`
    attribute of the params argument.
    """

    def __init__(self, archetype_resource_id) -> None:
        dir_path = os.path.join(ARCHETYPE_DIR, archetype_resource_id)
        self._archetype_resource_id = archetype_resource_id
        self._delegate = TemplateArchetype(dir_path)

    @property
    def archetype_resource_id(self) -> str:
        return self._archetype_resource_id

    @property
    def canonical_name(self) -> str:
        return self.name.lower()

    def file_paths(self, root_path: str, params: ArchetypeParameters) -> Iterable[str]:
        return self._delegate.file_paths(root_path, params)

    def dir_paths(self, root_path: str, params: ArchetypeParameters) -> Iterable[str]:
        return self._delegate.dir_paths(root_path, params)

    def build(self, root_dir: str, params: ArchetypeParameters) -> None:
        return self._delegate.build(root_dir, params)

    @classmethod
    def from_string(cls, s: str):
        return {sa.canonical_name: sa for sa in StandardArchetype}[s.lower()]

    @classmethod
    def canonical_names(cls) -> Tuple[str]:
        return tuple(sa.canonical_name for sa in StandardArchetype)



"""
archetype_descriptor
~~~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`ArchetypeDescriptor` along with supporting
classes, functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from collections import namedtuple

from jsonschema import validate, ValidationError

from inception_tools.json_serializable import (
    JSON_OBJ_TYPE,
    JsonSerializable,
    JsonSerializationError,
)


class _JsonKey(object):
    DIRECTORIES = "directories"
    FILES = "files"
    PROTOTYPE = "prototype"
    SUBPATH = "subpath"


DirectoryDescriptor = namedtuple("DirectoryDescriptor", ("subpath",))
FileDescriptor = namedtuple("FileDescriptor", ("subpath", "prototype"))


class ArchetypeDescriptor(
    namedtuple("_ArchetypeDescriptorBase", ("files", "directories")), JsonSerializable
):
    """
    Describes the files and directories which should be created by an
    :py:class:`Architype`.
    """

    # Make instances of this class immutable
    __slots__ = ()

    JSON_SCHEMA = {
        "type": "object",
        "properties": {
            "directories": {
                "type": "array",
                "items": {"$ref": "#/definitions/directory"},
            },
            "files": {"type": "array", "items": {"$ref": "#/definitions/file"}},
        },
        "definitions": {
            "directory": {
                "type": "object",
                "properties": {
                    "subpath": {"type": "string"},
                },
            },
            "file": {
                "type": "object",
                "properties": {
                    "subpath": {"type": "string"},
                    "prototype": {"type": "string"},
                },
            },
        },
    }
    """
    The Python `JSON schema`_ that :py:meth:`serialize_json` and
    :py:meth:`deserialize_json` must adhere to.

    .. _`JSON schema`: https://json-schema.org/
    """

    @classmethod
    def from_json(cls, json_obj: JSON_OBJ_TYPE):
        """
        Creates a new :py:class:`ArchetypeDescriptor` from a JSON-like Python object
        of the form:

        .. code-block::

            {
                'directories': [
                    {
                        'subpath': <subpath-jinja2-template-string>
                    },
                    ...
                    {
                        'subpath': <subpath-jinja2-template-string>
                    }
                ],
                'files': [
                    {
                        'subpath': <subpath-jinja2-template-string>,
                        'prototype': <subpath-jinja2-template-string>
                    },
                    ...
                    {
                        'subpath': <subpath-jinja2-template-string>,
                        'prototype': <subpath-jinja2-template-string>
                    }
                ]
            }
        """
        cls._validate_json(json_obj)

        if isinstance(json_obj, dict):

            d_json = json_obj[_JsonKey.DIRECTORIES]
            dir_descriptors = tuple(
                DirectoryDescriptor(j[_JsonKey.SUBPATH]) for j in d_json
            )

            f_json = json_obj[_JsonKey.FILES]
            file_descriptors = tuple(
                FileDescriptor(j[_JsonKey.SUBPATH], j[_JsonKey.PROTOTYPE])
                for j in f_json
            )

            return cls(file_descriptors, dir_descriptors)

        raise RuntimeError(
            f"Expected an object of type 'dict' but received: {json_obj!r}"
        )

    @classmethod
    def _validate_json(cls, json_obj: JSON_OBJ_TYPE):
        try:
            validate(json_obj, cls.JSON_SCHEMA)
        except ValidationError as e:
            to_raise = JsonSerializationError(f"Invalid JSON format: {json_obj}")
            raise to_raise from e

    def to_json(self) -> JSON_OBJ_TYPE:
        """
        Returns a JSON-like Python object of the form:

        .. code-block::

            {
                'directories': [
                    {
                        'subpath': <subpath-jinja2-template-string>
                    },
                    ...
                    {
                        'subpath': <subpath-jinja2-template-string>
                    }
                ],
                'files': [
                    {
                        'subpath': <subpath-jinja2-template-string>,
                        'prototype': <subpath-jinja2-template-string>
                    },
                    ...
                    {
                        'subpath': <subpath-jinja2-template-string>,
                        'prototype': <subpath-jinja2-template-string>
                    }
                ]
            }
        """
        # Suppressing protected member inspection because NamedTuple
        # documentation has _asdict as part of its API
        # noinspection PyProtectedMember
        return {
            _JsonKey.DIRECTORIES: [d._asdict() for d in self.directories],
            _JsonKey.FILES: [d._asdict() for d in self.files],
        }



#!/usr/bin/env python
"""
    setup.py
    ~~~~~~~~

    Package distribution file for the inception_tools library.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import os

import setuptools

if __name__ == "__main__":
    # allow setup.py to run from another directory
    dir_ = os.path.abspath(os.path.join(__file__, os.pardir))
    os.chdir(dir_)
    setuptools.setup()
